/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.api;

import java.util.Random;

import net.minecraft.block.Block;
import net.minecraft.block.BlockSapling;
import net.minecraft.block.state.IBlockState;
import net.minecraft.init.Blocks;
import net.minecraft.util.Rotation;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraft.world.gen.structure.StructureBoundingBox;
import net.minecraft.world.gen.structure.template.*;

import net.dries007.tfc.api.types.Tree;
import net.dries007.tfc.objects.blocks.BlocksTFC;
import net.dries007.tfc.objects.blocks.wood.BlockLeavesTFC;

public interface ITreeGenerator
{

    /**
     * This is a copy of the method included in the Template class, with some key differences.
     * This will ignore TEs / Entities, and does less checks for bad usage, since it will only be used for tree worldgen
     * It will do an additional check that the block is replaceable; important for tree growth; as to not replace other blocks
     *
     * @param worldIn     the world
     * @param pos         the position
     * @param template    the template
     * @param placementIn the placement settings
     */
    static void addStructureToWorld(World worldIn, BlockPos pos, Template template, PlacementSettings placementIn)
    {
        int flags = 2;
        ITemplateProcessor templateProcessor = new BlockRotationProcessor(pos, placementIn);
        StructureBoundingBox structureboundingbox = placementIn.func_186213_g();

        for (Template.BlockInfo template$blockinfo : template.field_186270_a)
        {
            BlockPos blockpos = Template.func_186266_a(placementIn, template$blockinfo.field_186242_a).func_177971_a(pos);
            Template.BlockInfo template$blockinfo1 = templateProcessor.func_189943_a(worldIn, blockpos, template$blockinfo);

            if (template$blockinfo1 != null)
            {
                Block block1 = template$blockinfo1.field_186243_b.func_177230_c();

                if ((!placementIn.func_186227_h() || block1 != Blocks.field_185779_df) && (structureboundingbox == null || structureboundingbox.func_175898_b(blockpos)))
                {
                    IBlockState iblockstate = template$blockinfo1.field_186243_b.func_185902_a(placementIn.func_186212_b());
                    IBlockState iblockstate1 = iblockstate.func_185907_a(placementIn.func_186215_c());

                    if (worldIn.func_180495_p(blockpos).func_185904_a().func_76222_j() || worldIn.func_180495_p(blockpos).func_177230_c() instanceof BlockLeavesTFC)
                        worldIn.func_180501_a(blockpos, iblockstate1, flags);

                }
            }
        }

        for (Template.BlockInfo template$blockinfo2 : template.field_186270_a)
        {
            BlockPos blockpos1 = Template.func_186266_a(placementIn, template$blockinfo2.field_186242_a).func_177971_a(pos);

            if (structureboundingbox == null || structureboundingbox.func_175898_b(blockpos1))
            {
                worldIn.func_175722_b(blockpos1, template$blockinfo2.field_186243_b.func_177230_c(), false);
            }

        }
    }

    /**
     * This only sets the properties used by ITreeGenerator.addStructureToWorld
     *
     * @return A default set of placement settings for tree generation
     */

    static PlacementSettings getDefaultSettings()
    {
        return new PlacementSettings().func_186226_b(false);
    }

    /**
     * This only sets the properties used by ITreeGenerator.addStructureToWorld
     *
     * @param rand For generating random settings
     * @return A set of placement settings with random rotation
     */
    static PlacementSettings getRandomSettings(Random rand)
    {
        return getDefaultSettings().func_186220_a(Rotation.values()[rand.nextInt(Rotation.values().length)]);
    }

    /**
     * Called to generate a tree. Each Tree must have one of these. Used for world gen and sapling growth
     *
     * @param manager an instance of the world's template manager. Used for getting structures.
     * @param world   The world
     * @param pos     The position where the sapling was / would've been
     * @param tree    The tree type to spawn
     * @param rand    A random to use in generation
     */
    void generateTree(TemplateManager manager, World world, BlockPos pos, Tree tree, Random rand);

    /**
     * Checks if a tree can be generated. This implementation checks height, radius, and light level
     *
     * @param world    The world
     * @param pos      The pos of the tree
     * @param treeType The tree type (for checking if the tree can generate)
     * @return true if the tree can generate.
     */
    default boolean canGenerateTree(World world, BlockPos pos, Tree treeType)
    {
        // Check if ground is flat enough
        final int radius = treeType.getMaxGrowthRadius();
        for (int x = -radius; x <= radius; x++)
        {
            for (int z = -radius; z <= radius; z++)
            {
                if ((x == 0 && z == 0) ||
                    world.func_180495_p(pos.func_177982_a(x, 0, z)).func_185904_a().func_76222_j() ||
                    ((x > 1 || z > 1) && world.func_180495_p(pos.func_177982_a(x, 1, z)).func_185904_a().func_76222_j()))
                    continue;
                return false;
            }
        }
        // Check if there is room directly upwards
        final int height = treeType.getMaxHeight();
        for (int y = 1; y <= height; y++)
            if (!world.func_180495_p(pos.func_177981_b(y)).func_185904_a().func_76222_j())
                return false;

        // Check if there is a solid block beneath
        if (!BlocksTFC.isSoil(world.func_180495_p(pos.func_177977_b())))
            return false;

        // Check the position for liquids, etc.
        if (world.func_180495_p(pos).func_185904_a().func_76224_d() || !world.func_180495_p(pos).func_185904_a().func_76222_j())
            if (!(world.func_180495_p(pos).func_177230_c() instanceof BlockSapling))
                return false;

        // Check if there is sufficient light level
        return world.func_175671_l(pos) >= 7;
    }
}
